import os
import re
import sys
import subprocess
from io import StringIO
from typing import Any, Generator
from rich.text import Text
from .pipeline import pipeline

ENCODING = os.environ.get('MOULTI_MANPAGE_ENCODING', 'utf-8')

def overtype_to_ansi(content: str) -> str:
	"""
	Under *BSD and macOS, man pages are formatted using overtypes, like an
	actual teleprinter would do.
	Rich/Textual handle ANSI escape sequences but not overtypes.
	"""
	# character, backspace, character = bold character
	content = re.sub(r'(.)\x08\1', '\x1b[1m\\1\x1b[0m', content)
	# underscore, backspace, character = underlined character
	content = re.sub(r'_\x08(.)', '\x1b[4m\\1\x1b[24m', content)
	return content

def dissect_manpage(filedesc: Any) -> dict:
	title = ''
	while line := filedesc.readline():
		line = line.strip('\n')
		if line:
			title = overtype_to_ansi(line)
			break
	all_lines = [overtype_to_ansi(line) for line in filedesc.readlines()]
	footer = all_lines.pop().strip('\n')

	sections: list[dict] = []
	current_section = {'title': '', 'text': ''}
	for line in all_lines:
		if line.startswith(' ') or line == '\n': # regular line
			current_section['text'] += line
		else: # section title
			current_section['text'] = current_section['text'].strip('\n')
			if current_section['title'] or current_section['text']:
				sections.append(current_section)
			current_section = {'title': line.strip('\n'), 'text': ''}

	return {
		'title': title,
		'sections': sections,
		'footer': footer,
	}

def line_indent_size(line: str) -> int:
	size = 0
	for char in line:
		if char != ' ':
			break
		size += 1
	return size

def section_indent_size(text: str) -> int|None:
	size = None
	for line in text.split('\n'):
		if not line:
			continue
		lis = line_indent_size(line)
		if size is None or lis < size:
			size = lis
	return size

def unindent(text: str, max_size: int = -1) -> str:
	size = section_indent_size(text)
	if size is None:
		return text
	if max_size >= 0:
		size = min(size, max_size)
	result = []
	for line in text.split('\n'):
		if line:
			line = line[size:]
		result.append(line)
	return '\n'.join(result)

def ansi_to_markup(text: str) -> str:
	return Text.from_ansi(text).markup

def commands(title: str, manpage: dict) -> Generator:
	"""
	Yield pipeline()-compatible triplets.
	"""
	if 'MOULTI_MANPAGE_NO_TITLE' not in os.environ:
		yield None, {'command': 'set', 'title': title}, None

	pid = os.getpid()
	counter = {'step': 1}
	def step(cmd: str, title: str, **kwargs: Any) -> tuple[str, dict, None]:
		step_id = f'manpage_{pid}_{counter["step"]}'
		counter['step'] += 1
		step = {'command': cmd, 'action': 'add', 'id': step_id, 'title': title, 'classes': 'manpage', **kwargs}
		return step_id, step, None

	yield step('divider', ansi_to_markup(manpage['title']), classes='manpage title')
	max_unindent = 3
	for section in manpage['sections']:
		text = unindent(section['text'], max_unindent)
		yield step('step', ansi_to_markup(section['title']), text=text, collapsed=False, max_height=0)
	yield step('divider', ansi_to_markup(manpage['footer']), classes='manpage footer')

def handle_manpage_from_file_descriptor(title: str, filedesc: Any) -> None:
	pipeline(commands(title, dissect_manpage(filedesc)))

def handle_manpage_from_string(title: str, data: str) -> None:
	handle_manpage_from_file_descriptor(title, StringIO(data))

def manpage_parse(args: dict) -> None:
	filepath = args['manpage_filepath']
	with open(filepath, 'r', encoding=ENCODING, errors='surrogateescape') as filedesc:
		title = f'{filepath}'
		handle_manpage_from_file_descriptor(title, filedesc)

def manpage_run(args: dict) -> None:
	command = args['command']
	os.environ['MAN_KEEP_FORMATTING'] = 'y'
	res = subprocess.run(command, check=False, stdout=subprocess.PIPE, encoding=ENCODING, errors='surrogateescape')
	if res.stdout:
		title = ' '.join(command)
		handle_manpage_from_string(title, res.stdout)
	if res.returncode and not res.stdout:
		sys.stderr.write(f'manpage_run: {command} produced no output and exited with status {res.returncode}\n')
		sys.exit(res.returncode)
	sys.exit(0)



from typing import Any
from textual.app import ComposeResult
from textual.events import Resize
from textual.containers import Horizontal
from textual.reactive import Reactive
from textual.widgets import Button, Static
from ..abstractquestion.tui import AbstractQuestion

DEFAULT_BUTTONS = [
	['yes', 'success', 'Yes'],
	['no', 'error', 'No'],
	['cancel', 'primary', 'Cancel'],
]

class ButtonQuestion(AbstractQuestion):
	"""
	This widget represents an interactive question in a script, program or process.
	"""

	BINDINGS = [
		("c", "to_clipboard(False)", "Copy"),
		("w", "to_clipboard(True)", "With question"),
	]

	known_width = Reactive(-1)

	def __init__(self, id: str, **kwargs: Any): # pylint: disable=redefined-builtin
		super().__init__(id=id, **kwargs)
		self.button_layout: list[int] = []
		self.button_sizes: list[int]|None = None
		self.regenerate_buttons()
		self.vertical = Static(classes='buttons')

	def regenerate_buttons(self) -> None:
		self.buttons = []
		disabled = self.answer is not None
		for value, classes, label in self.init_kwargs.get('button', DEFAULT_BUTTONS):
			self.buttons.append(Button(label, variant=classes, name=value, disabled=disabled))

	def compose_question(self) -> ComposeResult:
		with self.vertical:
			# Initial layout: all buttons in a single row:
			with Horizontal():
				yield from self.buttons

	def on_button_pressed(self, event: Button.Pressed) -> None:
		answer = event.button.name
		if answer is not None:
			self.got_answer(answer)

	def check_properties(self, kwargs: dict[str, Any]) -> None:
		super().check_properties(kwargs)
		for _, _, label in kwargs.get('button', []):
			self.check_markup(label)

	def export_properties(self) -> dict[str, Any]:
		prop = super().export_properties()
		# Buttons cannot be updated so we can simply export the value found in init_kwargs:
		prop['button'] = self.init_kwargs.get('button', DEFAULT_BUTTONS)
		return prop

	@AbstractQuestion.copy_to_clipboard
	def action_to_clipboard(self, with_question: bool = False) -> tuple[bool, str, str]:
		if self.answer is None:
			return False, '', 'Answer not available yet'
		data = f'Q: {self.question()}\nA: {self.answer}\n' if with_question else self.answer
		return True, data, ''

	def measure_buttons(self) -> None:
		if self.button_sizes is None:
			# Harvest current button sizes:
			self.button_sizes = [button.size.width + button.styles.margin.right for button in self.buttons]

	def compute_button_layout(self, width: int = -1) -> list[int]:
		"""
		Return a list of integers representing how many buttons each row should contain.
		"""
		width = self.query('Contents').first().size.width if width == -1 else width
		width -= self.buttons[0].styles.margin.left
		self.measure_buttons()
		rows = []
		row = 0
		row_width = width
		assert self.button_sizes
		for button_width in self.button_sizes:
			if button_width > row_width and row:
				rows.append(row)
				row = 0
				row_width = width
			row += 1
			row_width -= button_width
		rows.append(row)
		width = self.query('Contents').first().size.width
		return rows

	async def apply_button_layout(self, new_layout: list[int]) -> None:
		"""
		Given a list of integers representing how many buttons each row should contain, achieve that state by:
		- adding new Horizontal containers if needed
		- moving buttons
		- removing unnecessary Horizontal containers, if any
		"""
		horizontals = list(self.vertical.query('Horizontal').results())
		current_button_map: list[int] = []
		for horizontal_index, horizontal in enumerate(horizontals):
			current_button_map.extend(horizontal_index for _ in horizontal.query('Button').results())
		current_horizontal_count = max(current_button_map) + 1

		new_button_map: list[int] = []
		for horizontal_index, button_count in enumerate(new_layout):
			new_button_map.extend(horizontal_index for _ in range(button_count))
		new_horizontal_count = len(new_layout)

		buttons_to_move: dict[tuple[int, int], list[int]] = {}
		for button_index, current_horizontal_index in enumerate(current_button_map):
			new_horizontal_index = new_button_map[button_index]
			if current_horizontal_index != new_horizontal_index:
				move_direction = new_horizontal_index - current_horizontal_index
				move_direction //= abs(move_direction)
				buttons_to_move.setdefault((new_horizontal_index, move_direction), []).append(button_index)

		if not buttons_to_move:
			return

		# Add extra Horizontal containers if needed:
		if new_horizontal_count > current_horizontal_count:
			new_horizontals = list(Horizontal() for _ in range(current_horizontal_count, new_horizontal_count))
			horizontals.extend(new_horizontals)
			self.vertical.mount_all(new_horizontals)

		# Move buttons:
		disabled = self.answer is not None
		buttons_reference = self.init_kwargs.get('button', DEFAULT_BUTTONS)
		for movement, button_indices in buttons_to_move.items():
			fresh_buttons = []
			for button_index in button_indices:
				self.buttons[button_index].remove()
				button_value, button_classes, button_label = buttons_reference[button_index]
				fresh_button = Button(button_label, variant=button_classes, name=button_value, disabled=disabled)
				self.buttons[button_index] = fresh_button
				fresh_buttons.append(fresh_button)
			target_horizontal_index, direction = movement
			if direction > 0:
				horizontals[target_horizontal_index].mount_all(fresh_buttons, before=0)
			else:
				horizontals[target_horizontal_index].mount_all(fresh_buttons)

		# Remove unnecessary Horizontal containers, if any:
		if new_horizontal_count < current_horizontal_count:
			for horizontal_index in range(new_horizontal_count, current_horizontal_count):
				horizontals[horizontal_index].remove()

	async def watch_known_width(self, _old_width: int, _new_width: int) -> None:
		"""
		Compute and apply a new layout for buttons whenever the width changes.
		"""
		# Do nothing if we ignore our width:
		if self.known_width == -1:
			return
		# Do not bother for zero or one button:
		if len(self.buttons) <= 1:
			return
		# Do not bother if this widget is collapsed:
		if self.collapsible.collapsed:
			# Discard the known width to force a change when this widget is expanded again:
			self.known_width = -1
			return
		# Recompute the layout:
		new_layout = self.compute_button_layout()
		# Do not bother if the layout did not change:
		if new_layout == self.button_layout:
			return
		self.button_layout = new_layout
		await self.apply_button_layout(new_layout)

	def on_resize(self, event: Resize) -> None:
		if self.known_width != event.size.width:
			self.known_width = event.size.width

	DEFAULT_CSS = """
		Static.buttons {
			layout: vertical;
			&> Horizontal {
				align: center middle;
				height: 5; /* Button height + margin */
				&> Button { margin-right: 2; margin-left: 2; margin-top: 1; margin-bottom: 1; }
			}
		}
	"""
MoultiWidgetClass = ButtonQuestion



import os
import re
import sys
import pwd
import uuid
from struct import calcsize, unpack
from socket import socket as Socket, AF_UNIX, SOCK_STREAM, SOL_SOCKET
from socket import recv_fds as socket_recv_fds, send_fds
import json
from json.decoder import JSONDecodeError
from typing import Any, Callable

Message = dict[str, Any]
FDs = list[int]
TLVCallback = Callable[[Socket, str, str, bytes, FDs], None]
LogCallback = Callable[[str], None]

# recv_fds() uses CMSG_LEN instead of CMSG_SPACE to compute the ancillary data buffer size (socklen_t msg_controllen)
# and is thus liable to underestimate it. On some platforms (e.g. FreeBSD+amd64), this results in MSG_CTRUNC.
# A simple workaround is to adjust/increment max_fds.
CMSG_ADJUSTMENT = 1 if 'bsd' in sys.platform else 0

def recv_fds(sock: Socket, bufsize: int, maxfds: int, flags: int = 0) -> tuple[Any, Any, Any, Any]:
	return socket_recv_fds(sock, bufsize, maxfds + CMSG_ADJUSTMENT, flags)

def abstract_unix_sockets_supported() -> bool:
	# According to https://github.com/microsoft/WSL/issues/4240#issuecomment-620805115 :
	# > "Windows implementation of AF_UNIX does not support abstract sockets"
	# Consequently, assume only Linux supports abstract Unix sockets:
	return sys.platform == 'linux'

def run_dir() -> str:
	# This mess is *exactly* why we prefer abstract Unix sockets:
	if 'XDG_RUNTIME_DIR' in os.environ:
		return os.environ['XDG_RUNTIME_DIR']
	our_directory = 'moulti'
	if 'XDG_CACHE_HOME' in os.environ:
		return os.path.join(os.environ['XDG_CACHE_HOME'], our_directory)
	caches = os.path.expanduser('~/Library/Caches')
	if os.path.isdir(caches):
		return os.path.join(caches, our_directory)
	return os.path.join(os.path.expanduser('~/.cache'), our_directory)
	# Fallback on /tmp would have brought security implications (e.g. umask).

def current_username() -> str:
	uid = os.getuid()
	return pwd.getpwuid(uid).pw_name

def current_instance() -> str:
	return os.environ.get('MOULTI_INSTANCE') or 'default'

def make_socket_path(filename: str, abstract: bool, dirpath: str|None = None, printable: bool = True) -> str:
	if abstract:
		return ('@' if printable else '\0') + filename
	return os.path.join(dirpath or run_dir(), filename)

def default_moulti_socket_path() -> str:
	username = current_username()
	instance = current_instance()
	socket_filename = f'moulti-{username}-{instance}.socket'
	return make_socket_path(socket_filename, abstract_unix_sockets_supported())

def moulti_bind_path(abstract: bool) -> str:
	username = current_username()
	pid = os.getpid()
	socket_filename = f'moulti-client-{username}-{pid}.socket'
	return make_socket_path(socket_filename, abstract, printable=False)

def is_abstract_socket(socket_path: str) -> bool:
	return bool(socket_path and socket_path[0] == '\0')

def from_printable(socket_path: str) -> str:
	if socket_path and socket_path[0] == '@':
		socket_path = '\0' + socket_path[1:]
	return socket_path

def to_printable(socket_path: str) -> str:
	if is_abstract_socket(socket_path):
		socket_path = '@' + socket_path[1:]
	return socket_path

PRINTABLE_MOULTI_SOCKET = os.environ.get('MOULTI_SOCKET_PATH') or default_moulti_socket_path()
MOULTI_SOCKET = from_printable(PRINTABLE_MOULTI_SOCKET)

class MoultiProtocolException(Exception):
	pass

class MoultiConnectionClosedException(Exception):
	def __init__(self, expected_bytes: int = 0, anomaly: bool = True):
		self.expected_bytes = expected_bytes
		self.anomaly = anomaly
	def __str__(self) -> str:
		if self.anomaly:
			return f'connection closed by peer while expecting {self.expected_bytes} bytes'
		return 'connection closed by peer'

def moulti_unix_socket() -> Socket:
	return Socket(AF_UNIX, SOCK_STREAM, 0)

def moulti_listen(bind: str = MOULTI_SOCKET, backlog: int = 100, blocking: bool = False) -> tuple[Socket, bool]:
	try:
		server_socket = moulti_unix_socket()
		abstract = is_abstract_socket(bind)
		if not abstract:
			os.makedirs(os.path.dirname(bind), exist_ok=True)
		server_socket.bind(bind)
		server_socket.listen(backlog)
		server_socket.setblocking(blocking)
		return server_socket, abstract
	except Exception as exc:
		err = f'cannot listen on {to_printable(bind)} (with backlog={backlog} and blocking={blocking}): {exc}'
		raise MoultiProtocolException(err) from exc

def clean_socket(socket_path: str = PRINTABLE_MOULTI_SOCKET) -> None:
	if not socket_path.startswith('@'):
		try:
			os.unlink(socket_path)
		except FileNotFoundError:
			pass

def get_unix_credentials(socket: Socket) -> tuple[int, int, int]:
	# struct ucred is { pid_t, uid_t, gid_t }
	struct_ucred = '3i'
	from socket import SO_PEERCRED # pylint: disable=import-outside-toplevel
	unix_credentials = socket.getsockopt(SOL_SOCKET, SO_PEERCRED, calcsize(struct_ucred))
	pid, uid, gid = unpack(struct_ucred, unix_credentials)
	return pid, uid, gid

def moulti_connect(address: str = MOULTI_SOCKET, bind: str | None = None) -> Socket:
	client_socket = moulti_unix_socket()
	abstract = is_abstract_socket(address)
	final_bind = bind or moulti_bind_path(abstract)
	client_socket.bind(final_bind)
	if not is_abstract_socket(final_bind):
		# As a pure client, we do not actually need to keep our socket file:
		clean_socket(final_bind)
	client_socket.connect(address)
	return client_socket

def getraddr(socket: Socket) -> str:
	"""
	Return the remote address of the given socket.
	"""
	try:
		raddr = socket.getpeername()
	except OSError as ose:
		raddr = f'unknown/errno={ose.errno}'
	if not isinstance(raddr, str):
		raddr = raddr.decode('utf-8')
	raddr = raddr.replace('\0', '@')
	return raddr + ':fd' + str(socket.fileno())

# This is a simple text-based TLV (type-length-value) protocol.
# preamble examples: ":JSON:0000000002048:" ":TXT_:0000000002048:"
PREAMBLE_FIXED_LENGTH = 20
PREAMBLE_REGEX = r'^:[A-Z_]{4}:[0-9]{13}:$'

def parse_preamble(preamble_data: bytes) -> tuple[str, int]:
	"""
	Parse a Moulti TLV preamble; return type and length or raise
	MoultiProtocolException.
	"""
	try:
		preamble = preamble_data.decode('ascii')
	except UnicodeDecodeError as ude:
		raise MoultiProtocolException('non-ASCII preamble: {str(preamble)}') from ude
	if not re.match(PREAMBLE_REGEX, preamble):
		raise MoultiProtocolException(f'invalid preamble: {preamble}')
	data_type = preamble[1:5].rstrip('_')
	if not data_type:
		raise MoultiProtocolException(f'invalid data type: {preamble[1:5]}')
	data_length = int(preamble[6:19])
	return data_type, data_length

def read_fixed_amount_from_socket(socket: Socket, amount: int) -> bytes:
	data = bytes()
	while amount > 0:
		new_data = socket.recv(amount)
		if not new_data:
			raise MoultiConnectionClosedException(amount)
		data += new_data
		amount -= len(new_data)
	return data

def write_fixed_amount_to_socket(socket: Socket, data: bytes) -> None:
	if not data:
		return
	to_send = len(data)
	sent = 0
	while to_send > 0:
		just_sent = socket.send(data[sent:])
		to_send -= just_sent
		sent += just_sent

def read_tlv_data_from_socket(socket: Socket, max_fds: int = 1) -> tuple[str, bytes, FDs]:
	file_descriptors: FDs = []
	if max_fds > 0:
		_, file_descriptors, _, _ = recv_fds(socket, 0, max_fds)

	try:
		preamble = read_fixed_amount_from_socket(socket, PREAMBLE_FIXED_LENGTH)
	except MoultiConnectionClosedException as mcce:
		mcce.anomaly = mcce.expected_bytes != PREAMBLE_FIXED_LENGTH
		raise mcce
	data_type, data_length = parse_preamble(preamble)
	return data_type, read_fixed_amount_from_socket(socket, data_length), file_descriptors

def write_tlv_data_to_socket(socket: Socket, data: bytes, data_type: str = 'JSON', fds: FDs|None = None) -> None:
	data_type = data_type + '____'
	data_length = len(data)
	preamble = f':{data_type[0:4]}:{data_length:013}:'.encode('ascii')
	to_send = preamble + data
	if fds:
		send_fds(socket, [to_send], fds)
	else:
		write_fixed_amount_to_socket(socket, to_send)

def data_to_message(data: bytes) -> Message:
	try:
		string = data.decode('utf-8')
		message = json.loads(string)
		return message
	except UnicodeDecodeError as ude:
		raise MoultiProtocolException(f'received non UTF-8 {len(data)}-byte line') from ude
	except JSONDecodeError as jde:
		raise MoultiProtocolException('received non-JSON message') from jde

def message_to_data(message: Message) -> bytes:
	string = json.dumps(message)
	data = string.encode('utf-8')
	return data

def send_json_message(socket: Socket, message: Message, fds: FDs | None = None) -> str:
	# Setting a message id yields smaller replies as the server does not have to send the original data again:
	if 'msgid' not in message:
		message = {'msgid': str(uuid.uuid4()), **message}
	data = message_to_data(message)
	write_tlv_data_to_socket(socket, data, 'JSON', fds)
	return message['msgid']

def recv_json_message(socket: Socket, max_fds: int = 1) -> tuple[Message, FDs]:
	data_type, data, file_descriptors = read_tlv_data_from_socket(socket, max_fds)
	if data_type != 'JSON':
		raise MoultiProtocolException('received non-JSON message')
	return data_to_message(data), file_descriptors

def send_to_moulti(message: Message, wait_for_reply: bool = True) -> Message | None:
	with moulti_connect() as moulti_socket:
		msgid = send_json_message(moulti_socket, message)
		if wait_for_reply:
			reply, _ = recv_json_message(moulti_socket, 0)
			if msgid in reply and reply['msgid'] != msgid:
				raise MoultiProtocolException('expected message id {msgid} but received {reply["msgid"]}')
			return reply
	return None

class MoultiTLVReader:
	"""
	Helper that reads TLV data from a given socket each time read() is called.
	Each time a complete TLV dataset has been read, call a given callback.
	"""
	# pylint: disable=attribute-defined-outside-init
	def __init__(
		self,
		socket: Socket,
		raddr: str,
		callback: TLVCallback,
		log_callback: LogCallback|None = None,
		max_fds: int = 1
	):
		self.socket = socket
		self.raddr = raddr
		self.callback = callback
		self.log_callback = log_callback
		self.max_fds = max_fds
		self.reset()

	def reset(self) -> None:
		self.recv_fds_done = False
		self.file_descriptors: FDs = []

		self.recv_preamble_done = False
		self.preamble_data = bytes()
		self.data_type = ''
		self.data_length = 0
		self.data_value = bytes()

	def log(self, line: str) -> None:
		if self.log_callback:
			self.log_callback(line)

	def got_complete_tlv(self) -> None:
		data_type, data, file_descriptors = self.data_type, self.data_value, self.file_descriptors
		self.reset()
		self.callback(self.socket, self.raddr, data_type, data, file_descriptors)

	def read(self) -> None:
		try:
			# Read file descriptors:
			if self.max_fds > 0 and not self.recv_fds_done:
				_, self.file_descriptors, _, _ = recv_fds(self.socket, 0, self.max_fds)
				self.recv_fds_done = True
			# Read preamble:
			if not self.recv_preamble_done:
				amount = PREAMBLE_FIXED_LENGTH - len(self.preamble_data)
				while amount > 0:
					new_data = self.socket.recv(amount)
					if not new_data:
						# If we notice the client closed its connection here, it may be normal:
						anomaly = amount != PREAMBLE_FIXED_LENGTH
						raise MoultiConnectionClosedException(amount, anomaly=anomaly)
					self.preamble_data += new_data
					amount -= len(new_data)
				self.data_type, self.data_length = parse_preamble(self.preamble_data)
				self.recv_preamble_done = True
			# Read data:
			while self.data_length > 0:
				new_data = self.socket.recv(self.data_length)
				if not new_data:
					raise MoultiConnectionClosedException(self.data_length)
				self.data_value += new_data
				self.data_length -= len(new_data)
			self.got_complete_tlv()
		except BlockingIOError:
			# It turns out there is nothing (left) to read; better luck next read().
			return



#!/usr/bin/env python3

from setuptools import setup

short_description = 'Moulti is a CLI-driven Terminal User Interface (TUI) that enables you to assign the numerous '
short_description += 'lines emitted by your scripts to visual, collapsible blocks called steps.'

with open('README.md', 'r') as readme:
	long_description = readme.read()

setup(
	name='moulti',
	version='1.12.1',
	description=short_description,
	long_description=long_description,
	long_description_content_type='text/markdown',
	author='Xavier G.',
	author_email='xavier.moulti@kindwolf.org',
	url='https://github.com/xavierog/moulti',
	packages=[
		'moulti',
		'moulti.ansible',
		'moulti.widgets',
		'moulti.widgets.abstractstep',
		'moulti.widgets.divider',
		'moulti.widgets.collapsiblestep',
		'moulti.widgets.step',
		'moulti.widgets.abstractquestion',
		'moulti.widgets.buttonquestion',
		'moulti.widgets.inputquestion',
	],
	classifiers=[
		'Environment :: Console :: Curses',
		'Intended Audience :: Developers',
		'Intended Audience :: System Administrators',
		'License :: OSI Approved :: MIT License',
		'Operating System :: POSIX :: Linux',
		'Operating System :: MacOS :: MacOS X',
		'Programming Language :: Python :: 3 :: Only',
		'Topic :: System :: Logging',
		'Topic :: Utilities',
	],
	license='MIT',
	keywords=['cli', 'tui', 'curses', 'terminal', 'multiplex', 'script', 'output', 'steps', 'textual', 'collapsible', 'ansible', 'diff'],
	package_dir={'': 'src'},
	install_requires=['textual>=0.53', 'pyperclip', 'argcomplete', 'unidiff'],
	entry_points={
		'console_scripts': [
			'moulti = moulti.precli:main',
			'moulti-askpass = moulti.askpass:main',
		]
	},
	scripts=[
		'examples/moulti-functions.bash',
		'examples/moulti-man',
	],
)
